<?php
// $Id: ad_cache_file.inc,v 1.1.2.2 2007/05/03 14:20:25 jeremy Exp $

/**
 * @file
 * A plug in for the ad.module, providing a file cache mechanism for improved
 * performance when displaying ads.
 *
 * Copyright (c) 2007.
 *  Jeremy Andrews <jeremy@kerneltrap.org>.  All rights reserved.
 */

/**
 * This is the actual cache function called by adserve.php that displays ads
 * without bootstrapping Drupal.
 */
function ad_cache_file() {
  if (adserve_variable('debug')) {
    echo "Using file cache.<br />\n";
  }
  
  // We'll loop through all possible cache files until we obtain an 
  // exclusive lock.
  $lock = FALSE;
  for ($i = 1; $i <= adserve_variable('files'); $i++) {
    // Prefix the filename with a '.' to hide it on Unix systems.
    $cache_file = adserve_variable('path') .'/.'. $i .'.ad.cache';
  
    if (adserve_variable('debug')) {
      echo "Trying cache_file '$cache_file'.<br />\n";
    }
  
    if (!$fd = @fopen($cache_file, 'r+')) {
      if (adserve_variable('debug')) {
        echo "Failed to open cache_file '$cache_file'.<br />\n";
      }
      // We failed to open the cache file, try the next one.
      continue;
    }
    if ($i < adserve_variable('files')) {
      // This isn't the last available cache file so we'll use a 
      // non-blocking lock for best performance.  If we fail to lock this 
      // cache file, we'll quickly move on to the next until we find an 
      // available one.
      if (!flock($fd, LOCK_EX|LOCK_NB)) {
        if (adserve_variable('debug')) {
          echo "Failed to obtain non-blocking lock.<br />\n";
        }
        // We failed to obtain an exclusive lock, close the file and try the
        // next one.
        @fclose($fd);
        continue;
      }
      if (adserve_variable('debug')) {
        echo "Obtained lock.<br />\n";
      }
      $lock = TRUE;
      break;
    }
    else {
      // This is the last available cache file, we'll use a blocking lock 
      // as we have to wait until we have exclusive write permissions.
      if (!flock($fd, LOCK_EX)) {
        if (adserve_variable('debug')) {
          echo "Failed to obtain blocking lock.<br />\n";
        }
        // A blocking exclusive lock shouldn't ever fail, so something has
        // gone very wrong.  Perhaps the file was deleted out from under us?
        @fclose($fd);
        continue;
      }
      if (adserve_variable('debug')) {
        echo "Obtained lock.<br />\n";
      }
      $lock = TRUE;
      break;
    }
  }
   
  if ($lock) {
    // Read entire cache file into memory.
    $cache = unserialize(fread($fd, filesize($cache_file)));
    $group = adserve_variable('group');
     
    /**
     * The cache structure looks like this:
     * $cache['ad'][$group][]['display'] = $ad
     * $cache['ad'][$group][]['aid'] = $aid
     * $cache['ad'][$group][][$action][$timestamp] = $counter
     * $cache['last_sync'] = $timestamp
     */
     
    $output = '';
    $ids = array();
    for ($i = 0; $i < adserve_variable('quantity'); $i++) {
      if (adserve_variable('nids')) {
        $nids = explode(',', adserve_variable('nids'));
        $cache_size = sizeof($nids);
      }
      else {
        $cache_size = sizeof($cache['ad'][$group]);
      }
  
      if (adserve_variable('debug')) {
        echo "$i: cache_size($cache_size)<br />";
      }
  
      // We can't display more than $cache_size unique ads.
      if ($i >= $cache_size) 
        break;
      $id = FALSE;
  
      while ($id === FALSE || in_array($id, $ids)) {
        // Randomly select a unique banner to display.  We subtract 1 as 
        // arrays start at 0.
        $id = $cache_size > 1 ? mt_rand(0, $cache_size - 1) : 0;
        if (adserve_variable('debug')) {
          echo "Id: $id<br />";
        }
      }
  
      if (isset($nids)) {
        $group = FALSE;
        $loop = 0;
        while (!$group) {
          if (adserve_variable('debug')) {
            echo "$loop: random id: $id<br />";
          }
          $group = $cache['ad'][$nids[$id]]['group'];
          if (!$group) {
            // We were passed an invalid nid, walk through all available to
            // find one that is valid.  The ad that comes after a 
            // nonexistent ad will end up being favored by average, but so 
            // it goes...
            if ($id == $cache_size) {
              $id = 0;
            }
            else {
              $id++;
            }
            // If all passed in nid were invalid, prevent an infinite loop.
            if ($loop++ == $cache_size)
              break;
          }
        }
        $ad_id = $cache['ad'][$nids[$id]]['id'];
        $ids[] = $id;
        $id = $ad_id;
      }
      else {
        $ids[] = $id;
      }

      $ad = $cache['ad'][$group][$id];
      $last_sync = $cache['last_sync'];

      if ($output) {
        $output .= '<br />';
      }
      $output .= $ad['display'];
     
      if (adserve_variable('debug')) {
        echo "Cache size: $cache_size<br />\n";
        echo "Random ID: $id<br />\n";
        $aid = $cache['ad'][$group][$id]['aid'];
        echo "Ad ID: $aid<br />\n";
        echo "Last sync: $last_sync<br />\n";
      }

      // Increment view counter.
      $timestamp = date('YmdH');
      if (isset($cache['ad'][$group][$id]['view'])) {
        $cache['ad'][$group][$id]['view'][$timestamp]++;
      }
      else {
        $cache['ad'][$group][$id]['view'][$timestamp] = 1;
      }
    }

    // Write updated cache back to file and release the lock.
    $cache = serialize($cache);
    rewind($fd);
    fwrite($fd, $cache, strlen($cache));
    flock($fd, LOCK_UN);
    fclose($fd);
  
    // Every 60 seconds we flush the cache files to the database.
    if ($last_sync < time() - 60) {
      ad_cache_file_rebuild();
    }
  }
  else {
    $output = 'Failed to lock cache file.';
    if (ad_cache_file_rebuild()) {
      // Required function was missing, the file cache must be disabled, so 
      // return nothing allowing adserve.php to use the default display method.
      return;
    }
  }
  if (empty($output)) {
    $error = TRUE;
    if ($nids) {
      $output = 'No active ads were found in the nids "'. adserve_variable('nids') .'".';
    }
    else {
      $output = 'No active ads were found in the group "'. adserve_variable('group') .'".';
    }
  }
  return $output;
}

/**
 *  Additional variables required by the filecache.
 */
function ad_cache_file_variables() {
  // paths are comprised of alphanumerics, underscores, dashes, periods and 
  // slashes.
  $variables = array();
  $variables['path'] = isset($_GET['p']) ? preg_replace('/[^_\-\.\/\\0-9a-zA-Z]/', '', $_GET['p']) : '';
  // files is an integer.
  $variables['files'] = isset($_GET['f']) ? (int)$_GET['f'] : 0;
  if ($variables['files'] > 15) {
      echo "Invalid value 'f=". $variables['files'] ."', exiting.<br />\n";
      exit();
  }
  return $variables;
}

/**
 * Bootstrap drupal, then run _ad_build_cache_files() from ad.module which will 
 * rebuild all cache files.
 */
function ad_cache_file_rebuild() {
  adserve_bootstrap();
  if (function_exists('ad_cache_file_build')) {
    ad_cache_file_build();
    return 0;
  }
  else {
    return -1;
  }
}

?>
